from pyNN.nest import *

import time
import threading
import Queue

from trafficsignal.controller.EnumStates import *
from trafficsignal.model.Constants import *

# the main PyNN simulator model implementation as a threaded class

# It really has to be threaded - a couple of design/implmentation issues arose ...
# The PyNN similator simpy hogs resources, perhaps related to the Python GIL (global interpreter lock).
# There are also some thread implications with TKinter - and it proved impossible to drive the GUI
# from the model, it had to be the other way around i.e., starting the model from the GUI along with
# putting the PyNN model implmentation in a thread. In fact, my POSA book suggests that it is a better
# MVC design for the view to create the controller, which is what I have arrived at. In addition to this,
# this design needed to be combined with (threaded) publish/subscribe with a polling approach rather 
# than the simpler observer notification design

class NeuralModel:

    def __init__(self, queue):

        self.queue = queue

    def start(self):

        self.thread1 = threading.Thread(target=self.workerThread1)
        self.thread1.start(  )


    # the main work of the simulation

    def workerThread1(self):

        """
        This is where we handle the asynchronous I/O. One important thing to remember is that the thread has
        to yield control pretty regularly ... this is effectively done with sleeps (I think!).
        """

        self.setUpSimulation()

        run(SIMULATION_TIME, callbacks=[CheckState(400.0, self.pop_stop, self.pop_ready, self.pop_go, self.pop_stopping, self)])


    def setUpSimulation(self):

        setup(timestep=DELAY, min_delay=DELAY, max_delay=DELAY+99)

        cell_params = {'tau_refrac' : 5.0,'v_rest' : -65.0,'v_thresh' : -51.0,'tau_syn_E':2.0, 
            'tau_syn_I': 5.0,'v_reset': -70.0,'i_offset': 0.0,'cm': 0.1}
        spike1Times = [[1, 400, 800, 1200, 1600, 2000, 2400, 2800, 3200, 3600, 4000, 4400, 
                                4800, 5200, 5600, 6000, 6400, 6800, 7200, 7600, 8000]]
        spikeStart = [[1]]
        startSwitch = Population(1, SpikeSourceArray,{'spike_times': spikeStart})
        switch1 = Population(1, SpikeSourceArray,{'spike_times': spike1Times})

        self.pop_stop = Population(POPULATION_SIZE, IF_curr_exp, cell_params)
        self.pop_ready = Population(POPULATION_SIZE, IF_curr_exp, cell_params)
        self.pop_go = Population(POPULATION_SIZE, IF_curr_exp, cell_params)
        self.pop_stopping = Population(POPULATION_SIZE, IF_curr_exp, cell_params)

        Projection(self.pop_stop, self.pop_stop, AllToAllConnector(allow_self_connections=False),StaticSynapse(weight=WEIGHT))
        Projection(self.pop_ready,self.pop_ready,AllToAllConnector(allow_self_connections=False),StaticSynapse(weight=WEIGHT))
        Projection(self.pop_go,   self.pop_go,  AllToAllConnector(allow_self_connections=False),StaticSynapse(weight=WEIGHT))
        Projection(self.pop_stopping, self.pop_stopping, AllToAllConnector(allow_self_connections=False),StaticSynapse(weight=WEIGHT))

        injectionConnection = [(0, 0, 0.25*WEIGHT, DELAY)]

        Projection(switch1,self.pop_stop,FromListConnector(injectionConnection),StaticSynapse(weight=WEIGHT))
        Projection(switch1,self.pop_ready,FromListConnector(injectionConnection),StaticSynapse(weight=WEIGHT))
        Projection(switch1,self.pop_go,FromListConnector(injectionConnection),StaticSynapse(weight=WEIGHT))
        Projection(switch1,self.pop_stopping,FromListConnector(injectionConnection),StaticSynapse(weight=WEIGHT))

        startInjection = [(0, 0, 0.675*WEIGHT, DELAY)]

        Projection(startSwitch,self.pop_stop,FromListConnector(startInjection),StaticSynapse(weight=WEIGHT))

        onConnectors = []
        w=0.015*WEIGHT

        for i in range(0,POPULATION_SIZE,1):
            for j in range(0,POPULATION_SIZE,1):
                onConnectors=onConnectors+[(i,j,w,DELAY)]

        Projection(self.pop_stop,self.pop_ready, FromListConnector(onConnectors))
        Projection(self.pop_ready,self.pop_go,FromListConnector(onConnectors))
        Projection(self.pop_go,self.pop_stopping,FromListConnector(onConnectors))
        Projection(self.pop_stopping,self.pop_stop,FromListConnector(onConnectors))

        offConnectors = []
        w=-10*WEIGHT

        for i in range(0,POPULATION_SIZE,1):
            offConnectors=offConnectors+[(0,i,w,DELAY)]

        Projection(self.pop_ready,self.pop_stop,FromListConnector(offConnectors))
        Projection(self.pop_go,self.pop_ready,FromListConnector(offConnectors))
        Projection(self.pop_stopping,self.pop_go,FromListConnector(offConnectors))
        Projection(self.pop_stop,self.pop_stopping,FromListConnector(offConnectors))

        self.pop_stop.record(['spikes']) 
        self.pop_ready.record(['spikes'])
        self.pop_go.record(['spikes']) 
        self.pop_stopping.record(['spikes'])


    def mapEnumState(self, stop_value, ready_value, go_value, stopping_value):

        state_values = [stop_value, ready_value, go_value, stopping_value]
 
        index = 0
        for v in state_values:
            if v > MEAN_SPIKE_THRESHOLD:    # simplistic, but effective for this 'simple' simulation
                break
            else:
                index=index+1

        # switch on index
        if index == 0:
            self.current_enum_state = EnumStates.STOP
            self.queue.put(EnumStates.STOP)
            print("State switched to STOP")
            time.sleep(CLOCK_DELAY)
        elif index == 1:
            self.current_enum_state = EnumStates.READY
            self.queue.put(EnumStates.READY)
            print("State switched to READY")
            time.sleep(CLOCK_DELAY)
        elif index == 2:
            self.current_enum_state = EnumStates.GO
            self.queue.put(EnumStates.GO)
            print("State switched to GO")
            time.sleep(CLOCK_DELAY)
        elif index == 3:
            self.current_enum_state = EnumStates.STOPPING
            self.queue.put(EnumStates.STOPPING)
            print("State switched to STOPPING")
            time.sleep(CLOCK_DELAY)
        else:
            print("Error: Undetermined state")


#------------ PyNN simulation 'run' call back class -----------------

class CheckState:

    def __init__(self, interval, spop, rpop, gpop, stpop, parent):
        self.interval = interval
        self.stopped_pop = spop
        self.ready_pop = rpop
        self.go_pop = gpop
        self.stopping_pop = stpop
        self.parent = parent

    def __call__(self, t):

        print("Called at time: ", t) # so 'ideally' here inject pulse to all four populations and inspect data

        if t > 10.0: # i.e., don't try to grab data at time 0.0

            sc1 = self.stopped_pop.mean_spike_count()
            sc2 = self.ready_pop.mean_spike_count()
            sc3 = self.go_pop.mean_spike_count()
            sc4 = self.stopping_pop.mean_spike_count()

            print("mean spike counts (stop, ready, go stopping): ", sc1, sc2, sc3, sc4)

            # map and publish state
            self.parent.mapEnumState(sc1, sc2, sc3, sc4)

            # NOTE THAT THE MEAN VALUE (i.e., the actual number) VARIES WITH THE FREQUENCY OF CALL-BACKS ...
            # but the discrete states are consistent ...

            # effectively reset each call-back

            self.stopped_pop.get_data('spikes', clear=True)
            self.ready_pop.get_data('spikes', clear=True)
            self.go_pop.get_data('spikes', clear=True)
            self.stopping_pop.get_data('spikes', clear=True)

        return t + self.interval
