import time
import threading
import Queue

from trafficsignal.controller.EnumStates import *
from trafficsignal.model.PeriodicTimer import *

from abc import ABCMeta, abstractmethod

class ObjectModel:

    def __init__(self, queue):
        #self.queue = queue
        self.statemachine = ObjectStateMachine(queue)

    def start(self):
        self.runSimulation()

    def action(self):
        self.statemachine.moveState()

    # object specific implementation of abstract method
    def runSimulation(self):
        self.ptimer = PeriodicTimer(0.5)
        self.ptimer.start()
        threading.Thread(target=self.simulateWithThread, args=(20,)).start()

    # object specific
    def simulateWithThread(self, nticks):
        while nticks > 0:
            self.ptimer.wait_for_tick()
            self.action()
            nticks -= 1

class ObjectStateMachine:

    def __init__(self, queue):
        self.queue = queue
        self.currentState = StoppedState(self)

    def moveState(self):
        self.currentState.changeState() 
        #self.notifyObservers()
        # publish new state rather than notify
        
    # call-back for individual state objects to change object state
    def setState(self, newState):
        self.currentState = newState # object state
        self.mapEnumState()          # translate to enum state

    def mapEnumState(self):
        if isinstance(self.currentState, StoppedState):
            self.queue.put(EnumStates.STOP)
            print("State switched to STOP")
        if isinstance(self.currentState, ReadyState):
            self.queue.put(EnumStates.READY)
            print("State switched to READY")
        if isinstance(self.currentState, GoState):
            self.queue.put(EnumStates.GO)
            print("State switched to GO")
        if isinstance(self.currentState, StoppingState):
            self.queue.put(EnumStates.STOPPING)
            print("State switched to STOPPING")

#------------ BASE STATE CLASS -----------------

class TrafficState(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def changeState(self):
        pass

#----------Singleton Stopped State class--------

class StoppedState(TrafficState):
    __instance = None
    def __new__(cls, stateMachine):
        if StoppedState.__instance is None:
            StoppedState.__instance = object.__new__(cls)
        StoppedState.__instance.stateMachine = stateMachine
        return StoppedState.__instance

    def changeState(self):
        print('Changing from stopped to ready')
        self.stateMachine.setState(ReadyState(self.stateMachine))

    def __str__(self):
        return 'Stopped'
    

#----------Singleton Ready State class--------

class ReadyState(TrafficState):
    __instance = None
    def __new__(cls, stateMachine):
        if ReadyState.__instance is None:
            ReadyState.__instance = object.__new__(cls)
        ReadyState.__instance.stateMachine = stateMachine
        return ReadyState.__instance

    def changeState(self):
        print('Changing from ready to go')
        self.stateMachine.setState(GoState(self.stateMachine))

    def __str__(self):
        return 'Ready'


#----------Singleton Go State class--------

class GoState(TrafficState):
    __instance = None
    def __new__(cls, stateMachine):
        if GoState.__instance is None:
            GoState.__instance = object.__new__(cls)
        GoState.__instance.stateMachine = stateMachine
        return GoState.__instance

    def changeState(self):
        print('Changing from go to stopping')
        self.stateMachine.setState(StoppingState(self.stateMachine))

    def __str__(self):
        return 'Go'

#----------Singleton Stopping State class--------

class StoppingState(TrafficState):
    __instance = None
    def __new__(cls, stateMachine):
        if StoppingState.__instance is None:
            StoppingState.__instance = object.__new__(cls)
        StoppingState.__instance.stateMachine = stateMachine
        return StoppingState.__instance

    def changeState(self):
        print('Changing from stopping to stopped')
        self.stateMachine.setState(StoppedState(self.stateMachine))

    def __str__(self):
        return 'Stopping'
